@@ -246,6 +246,7 @@ static void scroll_off (ETree *et);
 static void scroll_on (ETree *et, guint scroll_direction);
 static void hover_off (ETree *et);
 static void hover_on (ETree *et, int x, int y);
+static void context_destroyed (gpointer data, GObject *ctx);
 
 static void
 et_disconnect_from_etta (ETree *et)
@@ -422,6 +423,10 @@ et_dispose (GObject *object)
 		if (et->priv->site)
 			e_tree_drag_source_unset (et);
 
+		if (et->priv->last_drop_context)
+			g_object_weak_unref (G_OBJECT(et->priv->last_drop_context), context_destroyed, et);
+		et->priv->last_drop_context = NULL;
+
 		gtk_widget_destroy (GTK_WIDGET (et->priv->table_canvas));
 
 		g_free(et->priv);
@@ -747,6 +752,9 @@ tree_canvas_size_allocate (GtkWidget *widget, GtkAllocation *alloc,
 	gdouble width;
 	gdouble height;
 	gdouble item_height;
+	GtkAdjustment *adj = GTK_LAYOUT(e_tree->priv->table_canvas)->vadjustment;
+	ETreePath path = e_tree_get_cursor (e_tree);
+	gint x, y, w, h;
 	GValue *val = g_new0 (GValue, 1);
 	g_value_init (val, G_TYPE_DOUBLE);
 
@@ -763,9 +771,18 @@ tree_canvas_size_allocate (GtkWidget *widget, GtkAllocation *alloc,
 		      NULL);
 	g_object_set_property (G_OBJECT (e_tree->priv->header), "width", val);
 	g_free (val);
+
 	if (e_tree->priv->reflow_idle_id)
 		g_source_remove(e_tree->priv->reflow_idle_id);
 	tree_canvas_reflow_idle(e_tree);
+
+	if (path)
+		e_tree_get_cell_geometry (e_tree, e_tree_row_of_node(e_tree, path), 0, &x, &y, &w, &h);
+	else
+		y = h = 0;
+
+ 	if (y < adj->value || y + h > adj->value + adj->page_size)
+		gtk_adjustment_set_value(adj, CLAMP(y - adj->page_size / 2, adj->lower, adj->upper - adj->page_size));
 }
 
 static void
@@ -2037,7 +2054,7 @@ find_next_in_range (ETree *et, gint start, gint end, ETreePathFunc func, gpointe
 
 	for (row = start; row <= end; row++) {
 		path = e_tree_table_adapter_node_at_row (et->priv->etta, row);
-		if (func (et->priv->model, path, data))
+		if (path && func (et->priv->model, path, data))
 			return path;
 	}
 
@@ -2052,7 +2069,7 @@ find_prev_in_range (ETree *et, gint start, gint end, ETreePathFunc func, gpointe
 
 	for (row = start; row >= end; row--) {
 		path = e_tree_table_adapter_node_at_row (et->priv->etta, row);
-		if (func (et->priv->model, path, data))
+		if (path && func (et->priv->model, path, data))
 			return path;
 	}
 
@@ -2716,7 +2733,6 @@ static gboolean
 hover_timeout (gpointer data)
 {
 	ETree *et = data;
-	GtkWidget *widget = data;
 	int x = et->priv->hover_x;
 	int y = et->priv->hover_y;
 	int row, col;
@@ -2795,7 +2811,7 @@ collapse_drag (ETree *et, ETreePath drop)
 }
 
 static void
-context_destroyed (gpointer data)
+context_destroyed (gpointer data, GObject *ctx)
 {
 	ETree *et = data;
 	if (et->priv) {
@@ -2813,10 +2829,15 @@ context_destroyed (gpointer data)
 static void
 context_connect (ETree *et, GdkDragContext *context)
 {
-	if (g_dataset_get_data (context, "e-tree") == NULL) {
+	if (context == et->priv->last_drop_context)
+       		return;
+	
+	if (et->priv->last_drop_context) 
+		g_object_weak_unref (G_OBJECT(et->priv->last_drop_context), context_destroyed, et);
+	else
 		g_object_ref (et);
-		g_dataset_set_data_full (context, "e-tree", et, context_destroyed);
-	}
+
+	g_object_weak_ref (G_OBJECT(context), context_destroyed, et);
 }
 
 static void
@@ -2853,8 +2874,8 @@ et_drag_motion(GtkWidget *widget,
 	et->priv->last_drop_x = x;
 	et->priv->last_drop_y = y;
 	et->priv->last_drop_time = time;
-	et->priv->last_drop_context = context;
 	context_connect (et, context);
+	et->priv->last_drop_context = context;
 
 	if (et->priv->hover_idle_id != 0) {
 		if (abs (et->priv->hover_x - x) > 3 ||
