@@ -68,27 +68,9 @@ color_combo_set_color_internal (ColorCombo *cc, GdkColor *color)
 			       NULL);
 }
 
-static void
-color_combo_destroy (GtkObject *object)
-{
-	GtkObjectClass *klass = (GtkObjectClass *)color_combo_parent_class;
-
-	if (klass->destroy)
-                klass->destroy (object);
-}
-
-static void
-color_combo_finalize (GObject *object)
-{
-	color_combo_parent_class->finalize (object);
-}
-
 static void
 color_combo_class_init (GObjectClass *object_class)
 {
-	object_class->finalize = color_combo_finalize;
-	((GtkObjectClass *)object_class)->destroy = color_combo_destroy;
-
 	color_combo_parent_class = g_type_class_ref (PARENT_TYPE);
 
 	color_combo_signals [CHANGED] =
